#!/usr/bin/perl

package Cognition::Export::Entity;
use strict;
use Cognition::RDFModel::Subject;
use Cognition::RDFModel::Property;
use Cognition::RDFModel::Value;
use Cognition::Export;
use Cognition::Export::Contact;
use Cognition::Export::Location;
use Cognition::Export::Calendar;
use Cognition::Export::Feed;
use Cognition::MagicString;
use HTML::Entities qw(encode_entities_numeric);
use JSON 2.0;
use URI;
use URI::Escape;

our $Encounters = {};

sub is
{
	my $subject = shift;
	return unless ($subject);
	return 1;
}

sub outputs
{
	return [
		{
			code  => 'RDF',
			name  => 'RDF/XML',
			type  => 'application/rdf+xml',
			multi => 1
		},
		{
			code  => 'JSON',
			name  => 'RDF/JSON',
			type  => 'application/x-rdf+json',
			multi => 1
		},
		{
			code  => 'Detect',
			name  => 'Detection Page',
			type  => 'text/html',
			multi => 1
		},
		{
			code  => 'Title',
			name  => 'Entity Title',
			type  => 'text/plain'
		}
	];
}


sub outTitle
{
	my $triples = $_[0];
	my $S       = $_[1];	
	my $subject = findSubject($triples, $S);

	my $t = firstOf($subject,
		'http://purl.org/dc/terms/title',
		'http://purl.org/dc/elements/1.1/title',
		'urn:ietf:rfc:2426#fn',
		'http://www.w3.org/2006/vcard/ns#fn',
		'http://www.w3.org/2006/03/hcard#fn',
		'http://www.w3.org/2001/vcard-rdf/3.0#FN',
		'http://xmlns.com/foaf/0.1/fullName',
		'http://xmlns.com/foaf/0.1/name',
		'http://www.w3.org/2000/10/swap/pim/contact#knownAs',
		'urn:ietf:rfc:2426#street-address',
		'urn:ietf:rfc:4287#title',
		'http://purl.org/rss/1.0/title',	
		'urn:ietf:rfc:2445#summary',
		'http://www.w3.org/2002/12/cal/ical#summary',
		'urn:ietf:rfc:2445#x-wr-calname');
	
	return $t if (length $t && $S =~ /^bnode\:\/\/\/(.*)$/);
	return "$t <$S>" if (length $t);
	return $1 if ($S =~ /^bnode\:\/\/\/(.*)$/);
	return $S;
}


sub multiDetect
{
	my $triples  = $_[0];
	my @subjects = sort SubjectSort (keys %$triples);
	
	my $rv = _outDetect($triples, '-');
	foreach my $s (@subjects)
	{
		$rv .= _outDetect($triples, $s);
	}
	return "<dl>\n$rv</dl>\n";
}


sub outDetect
{
	return "<dl>\n"._outDetect(@_)."</dl>\n";
}


sub _outDetect
{
	my $triples = $_[0];
	my $S       = $_[1];
	
	my $subject = findSubject($triples, $S);
	my $uri;

	if ($S eq '-')
	{
		my @S2 = keys %{$triples};
		my $S2 = shift @S2;
		$uri = $triples->{$S2}->{page}->abs_url;
	}
	else
	{
		$uri = $subject->{page}->abs_url;
		if ($uri.'#' eq substr($S, 0, length($uri)+1))
		{
			$uri = $S;
		}
		else
		{
			$uri .= "#subject($S)";
		}
	}
	$uri =~ s/^http\:\/\///;
	$uri = encode_entities_numeric(uri_escape(uri_escape($uri)));
	$uri =~ s/\%252F/\//ig;
	
	my $rv;
	if ($S eq '-')
		{ $rv = "\t<dt>(All on page)</dt>\n"; }
	else
		{ $rv = "\t<dt>".encode_entities_numeric(outTitle($triples, $S))."</dt>\n"; }
	
	foreach my $k (qw(Contact+vcard+vCard Contact+jcard+jCard
		Entity+rdf-xml+RDF/XML Entity+rdf-json+RDF/JSON Location+kml+KML
		Calendar+icalendar+iCalendar Feed+atom+Atom))
	{
		my ($kind, $path, $format) = split /\+/, $k;
		
		if (
			($S eq '-')
		||	($kind eq 'Contact' && Cognition::Export::Contact::is($subject))
		||	($kind eq 'Entity' && Cognition::Export::Entity::is($subject))
		||	($kind eq 'Location' && Cognition::Export::Location::is($subject))
		||	($kind eq 'Calendar' && Cognition::Export::Calendar::is($subject))
		||	($kind eq 'Feed' && Cognition::Export::Feed::is($subject))
		)
		{
			$rv .= "\t<dd><a href=\"/$path/$uri\">$kind as $format</a></dd>\n";
		}
	}
	
	return $rv;
}


sub multiJSON
{
	my $triples = shift;
	my %rv;
	foreach my $s (keys %$triples)
	{
		my $snice = $s;
		$snice = "_:$1" if ($s =~ m|^bnode:/*(.*)|);
		my $r   = _outJSON($triples, $s);
		$rv{$snice} = $r;
	}
 	return to_json(\%rv, {utf8 => 1, pretty => 1, canonical => 1, convert_blessed => 1})."\n";
}

sub outJSON
{
	my $triples = $_[0];
	my $S       = $_[1];
	my $json = _outJSON($triples, $S);
	my %rv;
	my $snice   = $_[1];
	$snice      = "_:$1" if ($_[1] =~ m|^bnode:/*(.*)|);
	$rv{$snice} = $json;

 	return to_json(\%rv, {utf8 => 1, pretty => 1, canonical => 1, convert_blessed => 1})."\n";
}

sub _outJSON
{
	my $triples = $_[0];
	my $S       = $_[1];
	my $subject = findSubject($triples, $S);
	my $rv      = {};

	foreach my $prop (values %{$subject->{properties}})
	{
		my @values;
		foreach my $val (@{$prop->valObjects})
		{
			next if ($prop->uri eq 'http://purl.org/dc/terms/identifier'
				&& $val->toString =~ /^bnode:/);

			my $V = {};
			$V->{type}  = $val->type ? 'uri' : 'literal';
			$V->{value} = $val->toString;
			if (isms($val->toString))
			{
				$V->{lang}  = $val->toString->{lang};
				$V->{value} = $val->toString->{string};
			}
			if ( $val->datatype )     { $V->{datatype} = $val->datatype; }
			
			if ($val->type && $val->toString =~ m|^bnode:/*(.*)|)
				{ $V->{value} = "_:$1"; }
			
			push @values, $V;
		}
		
		$rv->{ $prop->uri } = \@values if (@values);
	}
	
	return $rv;
}


sub _outRDF
{
	my $triples = $_[0];
	my $S       = $_[1];
	my $subject = findSubject($triples, $S);
	my $this    = $subject->{page};
	my $rv = '';
	
	$Cognition::Export::Entity::Encounters{$S} = 1;

	my $tag = firstOf($subject, 'http://www.w3.org/1999/02/22-rdf-syntax-ns#type');
	if ($tag)
		{ $tag = $this->_fq2pfx($tag); }
	else
		{ $tag = 'rdf:Description'; }

	$rv .= sprintf("\t<$tag%s>\n", rdf_about($triples, $subject->uri, 1));
	foreach my $property ($subject->propertiesObjects)
	{
		next if ($property->nVals == 0);
		next unless (length $property->namespace);
		my $pfxprop = $this->_fq2pfx($property->uri);
		
		foreach my $val (@{ $property->valObjects })
		{
			next if ($property->uri eq 'http://purl.org/dc/terms/identifier'
				&& $val->toString =~ /^bnode:/);
		
			my $langattr = '';
			if (isms($val->toString) && $val->toString->{lang})
				{ $langattr = sprintf(" xml:lang=\"%s\"", encode_entities_numeric($val->toString->{lang})); }

			my $datatypeattr = '';
			if ( $val->datatype )
				{ $datatypeattr = sprintf(" rdf:datatype=\"%s\"", encode_entities_numeric($val->datatype)); }
			
			my $xmlnsattr = '';
			my $valstr;
			if ($val->datatype eq 'http://www.w3.org/1999/02/22-rdf-syntax-ns#XMLLiteral' && !$val->type)
			{
				$valstr = $val->toString;
				$xmlnsattr = sprintf(" rdf:parseType=\"Literal\" xmlns=\"%s\"", encode_entities_numeric($val->{xmlns})); 
			}
			else
			{
				$valstr = encode_entities_numeric($val->toString); 
			}

			if ($val->type == 1)
			{
				if (defined $triples->{$valstr}
				&& $valstr =~ m#^(bnode:|tag:|http...openurl.invalid)#
				&& !$Cognition::Export::Entity::Encounters{$valstr})
				{
					my $nested = _outRDF($triples, $valstr);
					$nested =~ s/^/\t\t/gm;
					$rv .= sprintf("\t\t<%s>\n%s\t\t</%s>\n", $pfxprop, $nested, $pfxprop);
				}
				else
				{
					$rv .= sprintf("\t\t<%s%s />\n", $pfxprop, rdf_resource($triples, $valstr));
				}
			}
			elsif (length $val->vocab)
			{
				my $pfxvocab = $this->_fq2pfx($val->vocab);
			
				$rv .= sprintf("\t\t<%s>\n\t\t\t<%s>\n\t\t\t\t<rdf:value%s%s%s>%s</rdf:value>\n\t\t\t</%s>\n\t\t</%s>\n", 
									$pfxprop, $pfxvocab, $langattr, $datatypeattr, $xmlnsattr, $valstr, $pfxvocab, $pfxprop);
			}
			else
			{
				$rv .= sprintf("\t\t<%s%s%s%s>%s</%s>\n", 
									$pfxprop, $langattr, $datatypeattr, $xmlnsattr, $valstr, $pfxprop);
			}
		}
	}
	$rv .= "\t</$tag>\n";
	
	foreach my $alias (@{$subject->
		property('http://purl.org/dc/terms/identifier')
		->valObjects})
	{
		next unless ($alias->type);
		my $A = $alias->toString;
		$rv .= sprintf("\n\t<rdf:Description%s>"
			           ."\n\t\t<%s rdf:resource=\"%s\" />"
		              ."\n\t</rdf:Description>"
			           ."\n",
			           rdf_about($triples, $A, 1),
			           $this->_fq2pfx('http://www.w3.org/2002/07/owl#sameAs'),
			           encode_entities_numeric($subject->uri))
			unless ($A eq $S || $A =~ /\#xpointer\(/ || $A =~ /^bnode:/);
	}
	
	return $rv;
}


sub outRDF
{
	my $triples = $_[0];
	my $S       = $_[1];
	my $subject = $triples->{$S};
	my $this    = $subject->{page};

	$Cognition::Export::Entity::Encounters = {};

	my $rv = "<?xml version=\"1.0\"?>\n"
		. "<rdf:RDF";

	my (%used_namespaces);
	foreach my $p ( $subject->propertiesObjects )
		{ $used_namespaces{$p->namespace} = 1; }
	
	foreach my $ns (keys %{ $this->{NS} })
	{
		my $nsurl = $this->{NS}->{$ns}->{nsurl};
		next unless (length $nsurl);
		$rv .= "\n\txmlns:".lc($ns)."=\"".encode_entities_numeric($nsurl)."\""
			if ($used_namespaces{$nsurl});
	}
	$rv .= "NS_PLACEHOLDER_99754106633f94d350db34d548d6091a>\n\n";
	$rv .= _outRDF($triples, $S);
	$rv .= "</rdf:RDF>\n";
	return add_namespaces($rv, $triples, $S);
}

sub multiRDF
{
	my $triples = $_[0];
	my @keys = keys %$triples;
	my $key = shift @keys;
	my $this = $triples->{$key}->{page};
	
	$Cognition::Export::Entity::Encounters = {};

	my $rv = "<?xml version=\"1.0\"?>\n"
		. "<rdf:RDF";

	foreach my $ns (keys %{ $this->{NS} })
	{
		my $nsurl = $this->{NS}->{$ns}->{nsurl};
		next unless (length $nsurl);
		$rv .= "\n\txmlns:".lc($ns)."=\"".encode_entities_numeric($nsurl)."\"";
	}
	$rv .= "NS_PLACEHOLDER_99754106633f94d350db34d548d6091a>\n\n";
	
	my @subjects = keys %$triples;
	my @sortedSubjects = sort SubjectSort @subjects;
	foreach my $S (@sortedSubjects)
	{
		$rv .= _outRDF($triples, $S)."\n"
			unless ($Cognition::Export::Entity::Encounters{$S});
	}
	
	$rv .= "</rdf:RDF>\n";
	return add_namespaces($rv, $triples);
}

sub add_namespaces 
{
	my $string  = shift;
	my $triples = shift;
	my $subject = shift;
	my $rv = '';
	my %already;
	my $page;
	
	if ($subject)
	{
		my $S = findSubject($triples, $subject);
		$page = $S->{page};
		
		foreach my $P ( $S->propertiesObjects )
		{
			my $ns = $P->namespace;
			next unless (length $ns);
			$rv .= sprintf("\n xmlns:%s=\"%s\"", $page->{'NamespacePrefixes'}->{$ns}, $ns)
				unless ($already{$page->{'NamespacePrefixes'}->{$ns}});
			$already{$page->{'NamespacePrefixes'}->{$ns}} = $ns;
			
			# TODO
			# Also need to check within values for vocabs as these will be
			# namespaced. Plus add OWL.
		}
	}
	else
	{
		foreach my $subject (keys %$triples)
		{
			my $S = $triples->{$subject};
			$page = $S->{page};
			
			foreach my $P ($S->propertiesObjects)
			{
				my $ns = $P->namespace;
				next unless (length $ns);
				$rv .= sprintf("\n xmlns:%s=\"%s\"", $page->{'NamespacePrefixes'}->{$ns}, $ns)
					unless ($already{$page->{'NamespacePrefixes'}->{$ns}});
				$already{$page->{'NamespacePrefixes'}->{$ns}} = $ns;
			}
		}
	}
	
	$rv .= "\n xmlns:rdf=\"http://www.w3.org/1999/02/22-rdf-syntax-ns#\""
		unless ($already{$page->{'NamespacePrefixes'}
			->{'http://www.w3.org/1999/02/22-rdf-syntax-ns#'}});

	$string =~ s/NS_PLACEHOLDER_99754106633f94d350db34d548d6091a/$rv/;
	
	return $string;
}

sub rdf_resource
{
	my $triples = shift;
	my $uri     = shift;
	my $sure    = shift;
	
	unless ($sure)
	{
		my $subject = findSubject($triples, $uri);
		$uri = $subject->uri if ($subject);
	}
	
	if ($uri =~ m|^bnode:///(.*)$|)
		{ return sprintf(' rdf:nodeID="%s"', encode_entities_numeric($1)); }
		
	return sprintf(' rdf:resource="%s"', encode_entities_numeric($uri));
}

sub rdf_about
{
	my $triples = shift;
	my $uri     = shift;
	my $sure    = shift;
	
	unless ($sure)
	{
		my $subject = findSubject($triples, $uri);
		$uri = $subject->uri if ($subject);
	}
	
	if ($uri =~ m|^bnode:///(.*)$|)
		{ return sprintf(' rdf:nodeID="%s"', encode_entities_numeric($1)); }
		
	return sprintf(' rdf:about="%s"', encode_entities_numeric($uri));
}

sub SubjectSort
{
	return  1 if ($a =~ /^bnode/ && $b !~ /^bnode/);
	return -1 if ($b =~ /^bnode/ && $a !~ /^bnode/);
	
	return  1 if ($a =~ /\#/ && $b !~ /\#/);
	return -1 if ($b =~ /\#/ && $a !~ /\#/);
	
	$a cmp $b;
}

1;
